from __future__ import annotations

import time
from typing import TYPE_CHECKING

from qcportal.molecules import Molecule
from qcportal.record_models import RecordStatusEnum

if TYPE_CHECKING:
    from qcportal import PortalClient


def test_gridoptimization_full_1(fulltest_client: PortalClient):
    molecule = Molecule(
        symbols=["H", "O", "O", "H"],
        connectivity=[[0, 1, 1], [1, 2, 1], [2, 3, 1]],
        geometry=[
            1.848671612718783,
            1.4723466699847623,
            0.6446435664312682,
            1.3127881568370925,
            -0.1304193792618355,
            -0.2118922703584585,
            -1.3127927010942337,
            0.1334187339129038,
            -0.21189641512867613,
            -1.8386801669381663,
            -1.482348324549995,
            0.6446369709610646,
        ],
    )

    opt_spec = {
        "keywords": {},
        "program": "geometric",
        "protocols": {"trajectory": "none"},
        "qc_specification": {"basis": "sto-3g", "keywords": {}, "method": "blyp", "program": "psi4"},
    }

    go_keywords = {
        "preoptimization": False,
        "scans": [
            {"indices": [1, 2], "step_type": "relative", "steps": [-0.1, 0.0], "type": "distance"},
            {"indices": [0, 1, 2, 3], "step_type": "absolute", "steps": [-90, 0], "type": "dihedral"},
        ],
    }

    meta, ids = fulltest_client.add_gridoptimizations(
        initial_molecules=[molecule],
        program="gridoptimization",
        optimization_specification=opt_spec,
        keywords=go_keywords,
    )

    for i in range(240):
        time.sleep(1)
        rec = fulltest_client.get_gridoptimizations(ids[0])
        if rec.status not in [RecordStatusEnum.running, RecordStatusEnum.waiting]:
            break
    else:
        raise RuntimeError("Did not finish calculation in time")

    assert rec.status == RecordStatusEnum.complete


def test_gridoptimization_full_2(fulltest_client: PortalClient):
    molecule = Molecule(
        symbols=["H", "O", "O", "H"],
        connectivity=[[0, 1, 1], [1, 2, 1], [2, 3, 1]],
        geometry=[
            1.848671612718783,
            1.4723466699847623,
            0.6446435664312682,
            1.3127881568370925,
            -0.1304193792618355,
            -0.2118922703584585,
            -1.3127927010942337,
            0.1334187339129038,
            -0.21189641512867613,
            -1.8386801669381663,
            -1.482348324549995,
            0.6446369709610646,
        ],
    )

    input_opt_spec = {
        "keywords": {},
        "program": "geometric",
        "protocols": {"trajectory": "none"},
        "qc_specification": {"basis": "sto-3g", "keywords": {}, "method": "blyp", "program": "psi4"},
    }

    go_keywords = {
        "preoptimization": True,
        "scans": [
            {"indices": [1, 2], "step_type": "relative", "steps": [-0.1, 0.0], "type": "distance"},
            {"indices": [0, 1, 2, 3], "step_type": "absolute", "steps": [-90, 0], "type": "dihedral"},
        ],
    }

    meta, ids = fulltest_client.add_gridoptimizations(
        initial_molecules=[molecule],
        program="gridoptimization",
        optimization_specification=input_opt_spec,
        keywords=go_keywords,
    )

    for i in range(240):
        time.sleep(1)
        rec = fulltest_client.get_gridoptimizations(ids[0])
        if rec.status not in [RecordStatusEnum.running, RecordStatusEnum.waiting]:
            break
    else:
        raise RuntimeError("Did not finish calculation in time")

    assert rec.status == RecordStatusEnum.complete



from __future__ import annotations

import time
from typing import TYPE_CHECKING

from qcelemental.models import Molecule

from qcportal.record_models import RecordStatusEnum

if TYPE_CHECKING:
    from qcportal import PortalClient


def test_singlepoint_full_1(fulltest_client: PortalClient):
    mol = Molecule(symbols=["H", "H"], geometry=[[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]])
    meta, ids = fulltest_client.add_singlepoints(
        mol,
        "psi4",
        "energy",
        "hf",
        "sto-3g",
    )

    for i in range(120):
        time.sleep(1)
        rec = fulltest_client.get_singlepoints(ids[0])
        if rec.status not in [RecordStatusEnum.running, RecordStatusEnum.waiting]:
            break
    else:
        raise RuntimeError("Did not finish calculation in time")

    assert rec.status == RecordStatusEnum.complete


def test_singlepoint_full_2(fulltest_client: PortalClient):
    mol = Molecule(
        symbols=["H", "H", "O"],
        geometry=[[1.0, 1.0, 0.0], [-1.0, 1.0, 0.0], [0.0, 0.0, 0.0]],
        connectivity=[[0, 2, 1], [1, 2, 1]],
    )
    meta, ids = fulltest_client.add_singlepoints(
        mol,
        "rdkit",
        "energy",
        "uff",
        "",
    )

    for i in range(120):
        time.sleep(1)
        rec = fulltest_client.get_singlepoints(ids[0])
        if rec.status not in [RecordStatusEnum.running, RecordStatusEnum.waiting]:
            break
    else:
        raise RuntimeError("Did not finish calculation in time")

    assert rec.status == RecordStatusEnum.complete


def test_singlepoint_full_error_1(fulltest_client: PortalClient):
    # sto-3g not defined for U
    mol = Molecule(symbols=["U"], geometry=[[0.0, 0.0, 0.0]])

    meta, ids = fulltest_client.add_singlepoints(
        mol,
        "psi4",
        "energy",
        "b3lyp",
        "sto-3g",
    )

    for i in range(120):
        time.sleep(1)
        rec = fulltest_client.get_singlepoints(ids[0])
        if rec.status not in [RecordStatusEnum.running, RecordStatusEnum.waiting]:
            break
    else:
        raise RuntimeError("Did not finish calculation in time")

    assert rec.status == RecordStatusEnum.error
    assert "Unable to find a basis set" in rec.error["error_message"]


def test_singlepoint_full_error_2(fulltest_client: PortalClient):
    # rdkit requires connectivity
    mol = Molecule(symbols=["H", "H", "O"], geometry=[[1.0, 1.0, 0.0], [-1.0, 1.0, 0.0], [0.0, 0.0, 0.0]])
    meta, ids = fulltest_client.add_singlepoints(
        mol,
        "rdkit",
        "energy",
        "uff",
        "",
    )

    for i in range(120):
        time.sleep(1)
        rec = fulltest_client.get_singlepoints(ids[0])
        if rec.status not in [RecordStatusEnum.running, RecordStatusEnum.waiting]:
            break
    else:
        raise RuntimeError("Did not finish calculation in time")

    assert rec.status == RecordStatusEnum.error
    assert "RDKit requires molecules to have a connectivity graph" in rec.error["error_message"]



import json
import logging
import lzma
import sys

from qcarchivetesting.data_generator import DataGeneratorComputeThread
from qcfractal.snowflake import FractalSnowflake
from qcportal.molecules import Molecule
from qcportal.serialization import _JSONEncoder

logging.basicConfig(level=logging.WARNING)

if len(sys.argv) != 2:
    raise RuntimeError("Script takes a single argument - path to a test data input file")

infile_name = sys.argv[1]
outfile_name = infile_name + ".xz"

# Load the start of the test data
print(f"** Reading in data from {infile_name}")

with open(infile_name) as infile:
    test_data = json.load(infile)


# Set up the snowflake and compute process
print(f"** Starting snowflake")
snowflake = FractalSnowflake(compute_workers=0)
client = snowflake.client()
config = snowflake._qcf_config

# Add the data
_, ids = client.add_optimizations(
    [Molecule(**test_data["initial_molecule"])],
    program=test_data["specification"]["program"],
    qc_specification=test_data["specification"]["qc_specification"],
    keywords=test_data["specification"]["keywords"],
    protocols=test_data["specification"]["protocols"],
)

record_id = ids[0]

print(f"** Starting compute")
compute = DataGeneratorComputeThread(config, n_workers=1)

print(f"** Waiting for computation to finish")
snowflake.await_results([record_id], timeout=None)

print("** Computation complete. Assembling results **")
record = client.get_records(record_id)
if record.status != "complete":
    print(record.error)
    errs = client.query_records(status="error")
    for x in errs:
        print(x.error)
    raise RuntimeError(f"Record status is {record.status}")

result_data = compute.get_data()
assert len(result_data) == 1

task, result = result_data[0]
assert task.record_id == record_id
assert result["success"] is True

test_data["result"] = result

print(f"** Writing output to {outfile_name}")
with lzma.open(outfile_name, "wt") as f:
    json.dump(test_data, f, cls=_JSONEncoder, indent=4, sort_keys=True)

print(f"** Stopping compute worker")
compute.stop()



from __future__ import annotations

import logging
from copy import deepcopy

from qcarchivetesting import geoip_path, geoip_filename, ip_tests_enabled
from qcfractal.config import DatabaseConfig, update_nested_dict
from qcfractal.db_socket import SQLAlchemySocket
from qcfractal.postgres_harness import PostgresHarness, create_snowflake_postgres
from qcfractal.snowflake import FractalSnowflake
from qcportal import PortalClient, ManagerClient
from qcportal.auth import UserInfo, GroupInfo
from qcportal.managers import ManagerName
from .helpers import test_users, test_groups

_activated_manager_programs = {
    "qcengine": ["unknown"],
    "psi4": ["unknown"],
    "qchem": ["unknown"],
    "geometric": ["unknown"],
    "rdkit": ["unknown"],
    "mopac": ["unknown"],
    "prog1": ["unknown"],
    "prog2": ["unknown"],
    "prog3": ["unknown"],
    "prog4": ["unknown"],
    "optprog1": ["unknown"],
    "optprog2": ["unknown"],
    "optprog3": ["unknown"],
    "optprog4": ["unknown"],
}


class QCATestingPostgresHarness(PostgresHarness):
    def __init__(self, config: DatabaseConfig):
        PostgresHarness.__init__(self, config)
        self.db_name = self.config.database_name
        self.template_name = self.db_name + "_template"

    def create_template(self):
        """
        Creates a template database from an existing database

        It is expected that the template database does not already exist.
        """

        self.sql_command(
            f"CREATE DATABASE {self.template_name} TEMPLATE {self.db_name};",
            use_maintenance_db=True,
            returns=False,
        )

    def recreate_database(self):
        """
        Deletes the database, and recreates it from the template
        """

        self.delete_database()
        self.sql_command(
            f"CREATE DATABASE {self.db_name} TEMPLATE {self.template_name};",
            use_maintenance_db=True,
            returns=False,
        )


class QCATestingPostgresServer:
    """
    A temporary postgres instance used for testing
    """

    def __init__(self, db_path: str):
        self.logger = logging.getLogger(__name__)
        self.harness = create_snowflake_postgres(db_path)
        self.logger.debug(f"Using database located at {db_path} with uri {self.harness.config.safe_uri}")

        # Postgres process is up, but the database is not created
        assert self.harness.is_alive() and not self.harness.can_connect()

    def get_new_harness(self, db_name: str) -> QCATestingPostgresHarness:
        harness_config = deepcopy(self.harness.config.dict())
        harness_config["database_name"] = db_name

        new_harness = QCATestingPostgresHarness(DatabaseConfig(**harness_config))
        new_harness.create_database(create_tables=True)
        return new_harness


class QCATestingSnowflake(FractalSnowflake):
    """
    A snowflake class used for testing

    This mostly behaves like FractalSnowflake, but
    allows for some extra features such as manual handling of internal jobs
    and creating storage sockets from an instance.

    By default, the job runner and worker subprocesses are not started.
    """

    def __init__(
        self,
        pg_harness: QCATestingPostgresHarness,
        encoding: str,
        start_api=True,
        create_users=False,
        enable_security=False,
        allow_unauthenticated_read=False,
        log_access=True,
        extra_config=None,
    ):
        self.pg_harness = pg_harness
        self.encoding = encoding

        qcf_config = {}

        # Tighten the service frequency for tests
        # Also disable connection pooling in the storage socket
        # (which can leave db connections open, causing problems when we go to delete
        # the database)
        # Have a short token expiration (in case enable_security is True)

        # expire tokens in 5 seconds
        # Too short and this interferes with some other tests
        api_config = {"jwt_access_token_expires": 5}

        # Smaller api limits (so we can test chunking)
        api_limits = {
            "manager_tasks_claim": 5,
            "manager_tasks_return": 2,
            "get_records": 10,
            "get_dataset_entries": 5,
            "get_molecules": 11,
            "get_managers": 10,
            "get_server_stats": 10,
            "get_error_logs": 10,
            "get_access_logs": 10,
        }

        qcf_config["api"] = api_config
        qcf_config["api_limits"] = api_limits

        qcf_config["enable_security"] = enable_security
        qcf_config["allow_unauthenticated_read"] = allow_unauthenticated_read
        qcf_config["service_frequency"] = 5
        qcf_config["loglevel"] = "DEBUG"
        qcf_config["heartbeat_frequency"] = 3
        qcf_config["heartbeat_max_missed"] = 2
        qcf_config["statistics_frequency"] = 3

        qcf_config["database"] = {"pool_size": 0}
        qcf_config["log_access"] = log_access

        if ip_tests_enabled:
            qcf_config["geoip2_dir"] = geoip_path
            qcf_config["geoip2_filename"] = geoip_filename

        qcf_config["auto_reset"] = {"enabled": False}

        # Merge in any other specified config
        if extra_config:
            update_nested_dict(qcf_config, extra_config)

        FractalSnowflake.__init__(
            self,
            start=False,
            compute_workers=0,
            database_config=pg_harness.config,
            extra_config=qcf_config,
        )

        self._original_config = self._qcf_config.copy(deep=True)

        if create_users:
            self.create_users()

        # Start the flask api process if requested
        if start_api:
            self.start_api()

    def create_users(self):
        # Get a storage socket and add the roles/users/passwords
        storage = self.get_storage_socket()

        for g in test_groups:
            storage.groups.add(GroupInfo(groupname=g))

        for k, v in test_users.items():
            uinfo = UserInfo(username=k, enabled=True, **v["info"])
            storage.users.add(uinfo, password=v["pw"])

    def reset(self):
        self._stop_job_runner()
        self._stop_compute()
        self._all_completed = set()
        self._qcf_config = self._original_config.copy(deep=True)
        self.pg_harness.recreate_database()

    def get_storage_socket(self) -> SQLAlchemySocket:
        """
        Obtain a new SQLAlchemy socket

        This function will create a new socket instance every time it is called
        """

        return SQLAlchemySocket(self._qcf_config)

    def activate_manager(self):
        """
        Activates a manager on the server, but does not start any manager compute process
        """

        mname = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")
        socket = self.get_storage_socket()

        mid = socket.managers.activate(
            name_data=mname,
            manager_version="v2.0",
            username="bill",
            programs=_activated_manager_programs,
            tags=["*"],
        )

        return mname, mid

    def activated_manager_programs(self):
        return _activated_manager_programs

    def start_api(self) -> None:
        """
        Starts the flask/api thread
        """
        self._start_api()

    def stop_api(self) -> None:
        """
        Stops the flask/api thread
        """
        self._stop_api()

    def start_job_runner(self) -> None:
        """
        Starts the job runner thread
        """
        self._start_job_runner()

    def stop_job_runner(self) -> None:
        """
        Stops the job_runner thread
        """
        self._stop_job_runner()

    def client(self, username=None, password=None, cache_dir=None) -> PortalClient:
        """
        Obtain a client connected to this snowflake

        Parameters
        ----------
        username
            The username to connect as
        password
            The password to use
        cache_dir
            Directory to store cache files in

        Returns
        -------
        :
            A PortalClient that is connected to this snowflake
        """

        client = PortalClient(self.get_uri(), username=username, password=password, cache_dir=cache_dir)
        client.encoding = self.encoding
        return client

    def manager_client(self, name_data: ManagerName, username=None, password=None) -> ManagerClient:
        """
        Obtain a manager client connected to this snowflake

        Parameters
        ----------
        username
            The username to connect as
        password
            The password to use

        Returns
        -------
        :
            A PortalClient that is connected to this snowflake
        """

        # Now that we know it's up, create a manager client
        client = ManagerClient(name_data, self.get_uri(), username=username, password=password)
        return client



from __future__ import annotations

import time
from typing import TYPE_CHECKING

from qcportal.neb import NEBKeywords
from qcportal.optimization import OptimizationSpecification
from qcportal.record_models import RecordStatusEnum
from qcarchivetesting import load_molecule_data
from qcportal.singlepoint import QCSpecification

if TYPE_CHECKING:
    from qcportal import PortalClient


def test_neb_full_1(fulltest_client: PortalClient):
    """
    Full NEB test with aligning images.
    """
    chain = [load_molecule_data("neb/neb_HCN_%i" % i) for i in range(11)]

    # "maximum_force" and "average_force" are set to be high in orger to converge quickly.
    neb_keywords = NEBKeywords(
        images=7,
        spring_constant=1,
        spring_type=0,
        maximum_force=3,
        average_force=2,
        maximum_cycle=100,
        optimize_ts=False,
        optimize_endpoints=False,
        align=True,
        epsilon=1e-6,
        hessian_reset=True,
    )

    sp_spec = QCSpecification(
        program="psi4",
        driver="gradient",
        method="hf",
        basis="3-21g",
        keywords={},
    )

    meta, ids = fulltest_client.add_nebs(
        initial_chains=[chain],
        program="geometric",
        singlepoint_specification=sp_spec,
        optimization_specification=None,
        keywords=neb_keywords,
    )

    for i in range(600):
        time.sleep(15)
        rec = fulltest_client.get_nebs(ids[0])
        if rec.status not in [RecordStatusEnum.running, RecordStatusEnum.waiting]:
            break
    else:
        raise RuntimeError("Did not finish calculation in time")

    ts_guess = rec.result
    initial_chain = rec.initial_chain  # List[Molecule]
    final_chain = rec.final_chain  # List[Singlepoints]
    optimizations = rec.optimizations
    singlepoints = rec.singlepoints
    ts_optimization = rec.ts_optimization
    hessian = rec.ts_hessian

    # Finding the highest energy image from the last iteration SinglepointRecords.
    last_sps = singlepoints[max(singlepoints.keys())]
    sp_energies = [(sp.properties["current energy"], sp.molecule_id) for sp in last_sps]
    energy, neb_result_id = max(sp_energies)

    # Completed?
    assert rec.status == RecordStatusEnum.complete
    # Initial chain length and number of singlepoint records from the last iteration should be 11.
    assert len(initial_chain) == 7 and len(initial_chain) == len(final_chain)
    # SinglepointRecords ids of final chain should be the same as the last iteration SinglepointRecords from rec.singlepoints.
    assert sum([1 if i.id == j.id else 0 for i, j in zip(final_chain, singlepoints[max(singlepoints.keys())])]) == 7
    # There shouldn't be any OptimizationRecords.
    assert len(optimizations) == 0
    # There shouldn't be ts_optimization record.
    assert ts_optimization is None
    # When optimize_ts is False, there should not ba a record for the Hessian.
    assert hessian is None
    # SinglepointRecords should not have 'return_hessian' either.
    assert sum(1 if singlepoints[0][i].properties["return_hessian"] is None else 0 for i in range(7)) == 7
    # Result of the neb and the highest energy image of the last iteration should have the same molecule id.
    assert ts_guess.id == neb_result_id


def test_neb_full_2(fulltest_client: PortalClient):
    """
    Full NEB test without aligning images, with endpoint optimizations and TS optimization.
    This test takes longer than the previous one.
    """
    chain = [load_molecule_data("neb/neb_HCN_%i" % i) for i in range(11)]

    neb_keywords = NEBKeywords(
        images=11,
        spring_constant=1,
        spring_type=0,
        maximum_force=0.05,
        average_force=0.025,
        maximum_cycle=100,
        optimize_ts=True,
        optimize_endpoints=True,
        align=False,
        epsilon=1e-6,
        hessian_reset=True,
    )

    sp_spec = QCSpecification(
        program="psi4",
        driver="gradient",
        method="hf",
        basis="3-21g",
        keywords={},
    )

    opt_spec = OptimizationSpecification(
        program="geometric",
        qc_specification=sp_spec,
    )

    meta, ids = fulltest_client.add_nebs(
        initial_chains=[chain],
        program="geometric",
        singlepoint_specification=sp_spec,
        optimization_specification=opt_spec,
        keywords=neb_keywords,
    )

    for i in range(600):
        time.sleep(15)
        rec = fulltest_client.get_nebs(ids[0])
        if rec.status not in [RecordStatusEnum.running, RecordStatusEnum.waiting]:
            break
    else:
        raise RuntimeError("Did not finish calculation in time")

    ts_guess = rec.result
    initial_chain = rec.initial_chain  # List[Molecule]
    final_chain = rec.final_chain  # List[Singlepoints]
    optimizations = rec.optimizations
    singlepoints = rec.singlepoints
    ts_optimization = rec.ts_optimization
    hessian = rec.ts_hessian

    # Finding the highest energy image from the last iteration SinglepointRecords.
    last_sps = singlepoints[max(singlepoints.keys())]
    sp_energies = [(sp.properties["current energy"], sp.molecule_id) for sp in last_sps]
    energy, neb_result_id = max(sp_energies)

    # Completed?
    assert rec.status == RecordStatusEnum.complete
    # Initial chain length and number of singlepoint records from the last iteration should be 11.
    assert len(initial_chain) == 11 and len(initial_chain) == len(final_chain)
    # SinglepointRecords ids of final chain should be the same as the last iteration SinglepointRecords from rec.singlepoints.
    assert sum([1 if i.id == j.id else 0 for i, j in zip(final_chain, singlepoints[max(singlepoints.keys())])]) == 11
    # Total 3 OptimizationRecords: two for the endpoints, one for the TS optimization.
    assert len(optimizations) == 3
    # rec.tsoptimization should have the same id as the transition record in rec.optimizations.
    assert optimizations.get("transition").id == ts_optimization.id
    # When optimize_ts is True, SinglepointRecord containing the Hessian should exist.
    assert hessian
    # The rec.hessian should have the Hessian used for the TS optimization.
    assert hessian.properties["return_hessian"] is not None
    # Other SinglepointRecords should not have 'return_hessian'
    assert sum(1 if singlepoints[0][i].properties["return_hessian"] is None else 0 for i in range(11)) == 11
    # Result of the neb and the highest energy image of the last iteration should have the same molecule id.
    assert ts_guess.id == neb_result_id



"""
Contains testing infrastructure for QCFractal.
"""
from __future__ import annotations

import json
import lzma
import os
import signal
import sys
import time
from contextlib import contextmanager

from qcelemental.models import Molecule
from qcelemental.models.results import WavefunctionProperties

from qcfractal.components.serverinfo.socket import geoip2_found
from qcportal.serialization import _json_decode

# Valid client encodings
valid_encodings = ["application/json", "application/msgpack"]

# Path to this file (directory only)
_my_path = os.path.dirname(os.path.abspath(__file__))

geoip_path = os.path.join(_my_path, "MaxMind-DB", "test-data")
geoip_filename = "GeoLite2-City-Test.mmdb"
ip_testdata_path = os.path.join(_my_path, "MaxMind-DB", "source-data", "GeoIP2-City-Test.json")

ip_tests_enabled = os.path.exists(geoip_path) and os.path.exists(ip_testdata_path) and geoip2_found

testconfig_path = os.path.join(_my_path, "config_files")
migrationdata_path = os.path.join(_my_path, "migration_data")

test_groups = ["group1", "group2", "group3"]
test_users = {
    "admin_user": {
        "pw": "something123",
        "info": {
            "role": "admin",
            "groups": ["group1", "group2"],
            "fullname": "Mrs. Admin User",
            "organization": "QCF Testing",
            "email": "admin@example.com",
        },
    },
    "read_user": {
        "pw": "something123",
        "info": {
            "role": "read",
            "groups": ["group2"],
            "fullname": "Mr. Read User",
            "organization": "QCF Testing",
            "email": "read@example.com",
        },
    },
    "monitor_user": {
        "pw": "something123",
        "info": {
            "role": "monitor",
            "groups": ["group1"],
            "fullname": "Mr. Monitor User",
            "organization": "QCF Testing",
            "email": "monitor@example.com",
        },
    },
    "compute_user": {
        "pw": "something123",
        "info": {
            "role": "compute",
            "groups": [],
            "fullname": "Mr. Compute User",
            "organization": "QCF Testing",
            "email": "compute@example.com",
        },
    },
    "submit_user": {
        "pw": "something123",
        "info": {
            "role": "submit",
            "groups": ["group1"],
            "fullname": "Mrs. Submit User",
            "organization": "QCF Testing",
            "email": "submit@example.com",
        },
    },
}


def read_record_data(name: str):
    """
    Loads pre-computed/dummy procedure data from the test directory

    Parameters
    ----------
    name
        The name of the file to load (without the json extension)

    Returns
    -------
    :
        A dictionary with all the data in the file

    """

    data_path = os.path.join(_my_path, "procedure_data")
    file_path = os.path.join(data_path, name + ".json.xz")
    is_xz = True

    if not os.path.exists(file_path):
        file_path = os.path.join(data_path, name + ".json")
        is_xz = False

    if not os.path.exists(file_path):
        raise RuntimeError(f"Procedure data file {file_path} not found!")

    if is_xz:
        with lzma.open(file_path, "rt") as f:
            data = json.load(f, object_hook=_json_decode)
    else:
        with open(file_path, "r") as f:
            data = json.load(f, object_hook=_json_decode)

    return data


def load_molecule_data(name: str) -> Molecule:
    """
    Loads a molecule object for use in testing
    """

    data_path = os.path.join(_my_path, "molecule_data")
    file_path = os.path.join(data_path, name + ".json")
    return Molecule.from_file(file_path)


def load_wavefunction_data(name: str) -> WavefunctionProperties:
    """
    Loads a wavefunction object for use in testing
    """

    data_path = os.path.join(_my_path, "wavefunction_data")
    file_path = os.path.join(data_path, name + ".json")

    with open(file_path, "r") as f:
        data = json.load(f)
    return WavefunctionProperties(**data)


def load_ip_test_data():
    """
    Loads data for testing IP logging
    """

    with open(ip_testdata_path, "r") as f:
        d = json.load(f)

    # Stored as a list containing a dictionary with one key. Convert to a regular dict
    ret = {}
    for x in d:
        ret.update(x)

    return ret


def load_hash_test_data():
    """
    Loads data for testing dictionary hashing
    """

    file_path = os.path.join(_my_path, "hash_data", "dict_hash_test_data.json.xz")
    with lzma.open(file_path, "rt") as f:
        return json.load(f)


@contextmanager
def caplog_handler_at_level(caplog_fixture, level, logger=None):
    """
    Helper function to set the caplog fixture's handler to a certain level as well, otherwise it wont be captured

    e.g. if caplog.set_level(logging.INFO) but caplog.handler is at logging.CRITICAL, anything below CRITICAL wont be
    captured.
    """
    starting_handler_level = caplog_fixture.handler.level
    caplog_fixture.handler.setLevel(level)
    with caplog_fixture.at_level(level, logger=logger):
        yield
    caplog_fixture.handler.setLevel(starting_handler_level)


def terminate_process(proc):
    if proc.poll() is None:
        # Interrupt (SIGINT)
        if sys.platform.startswith("win"):
            proc.send_signal(signal.CTRL_BREAK_EVENT)
        else:
            proc.send_signal(signal.SIGINT)

        try:
            start = time.time()
            while (proc.poll() is None) and (time.time() < (start + 15)):
                time.sleep(0.02)

        # Kill (SIGKILL)
        finally:
            proc.kill()



from __future__ import annotations

import queue
import tempfile
import threading
import weakref
from typing import TYPE_CHECKING

from qcfractalcompute import ComputeManager
from qcfractalcompute.config import FractalComputeConfig, FractalServerSettings, LocalExecutorConfig

if TYPE_CHECKING:
    from typing import List, Dict, Any
    from qcportal.record_models import RecordTask
    from qcfractal.config import FractalConfig
    from qcfractalcompute.apps.models import AppTaskResult


def clean_conda_env(d: Dict[str, Any]):
    """
    Remove the bulky conda env info from provenances
    """
    if "conda_environment" in d:
        del d["conda_environment"]

    for v in d.values():
        if isinstance(v, dict):
            clean_conda_env(v)
        elif isinstance(v, list):
            for x in v:
                if isinstance(x, dict):
                    clean_conda_env(x)


class DataGeneratorManager(ComputeManager):
    def __init__(self, qcf_config: FractalConfig, result_queue: queue.Queue, n_workers: int = 2):
        self._qcf_config = qcf_config
        self._result_queue = result_queue
        self._record_id_map = {}  # Maps task id to record id

        # Maps task id to full task
        self._task_map = {}

        host = self._qcf_config.api.host
        port = self._qcf_config.api.port
        uri = f"http://{host}:{port}"

        tmpdir = tempfile.TemporaryDirectory()
        self._compute_config = FractalComputeConfig(
            base_folder=tmpdir.name,
            cluster="datagenerator_compute",
            update_frequency=10,
            server=FractalServerSettings(
                fractal_uri=uri,
                verify=False,
            ),
            executors={
                "local": LocalExecutorConfig(
                    scratch_directory=tmpdir.name,
                    cores_per_worker=1,
                    memory_per_worker=1,
                    max_workers=n_workers,
                    queue_tags=["*"],
                )
            },
        )
        ComputeManager.__init__(self, self._compute_config)

        def cleanup(d):
            d.cleanup()

        weakref.finalize(self, cleanup, tmpdir)

    def postprocess_results(self, results: Dict[int, AppTaskResult]):
        for task_id, app_result in results.items():
            # Return full task + full result (as dict)
            r_dict = app_result.result
            clean_conda_env(r_dict)
            self._result_queue.put((self._task_map[task_id], r_dict))

    def preprocess_new_tasks(self, new_tasks: List[RecordTask]):
        for task in new_tasks:
            # Store the full task by task id
            self._task_map[task.id] = task


class DataGeneratorComputeThread:
    def __init__(self, qcf_config: FractalConfig, n_workers: int = 2):
        self._qcf_config = qcf_config

        self._result_queue = queue.Queue()

        self._compute = DataGeneratorManager(self._qcf_config, self._result_queue, n_workers)
        self._compute_thread = threading.Thread(target=self._compute.start, daemon=True)

        self._finalizer = weakref.finalize(
            self,
            self._stop,
            self._compute,
            self._compute_thread,
        )

        self._compute_thread.start()

    @classmethod
    def _stop(cls, compute, compute_thread):
        if compute is not None:
            compute.stop()
            compute_thread.join()

    def stop(self) -> None:
        self._finalizer()

    def get_data(self) -> List[tuple[RecordTask, Dict[str, Any]]]:
        # Returns list of iterable (task, result)
        data = []

        while not self._result_queue.empty():
            d = self._result_queue.get(False)
            data.append(d)

        return data



from importlib.metadata import version

__version__ = version("qcarchivetesting")

from .helpers import (
    geoip_path,
    geoip_filename,
    ip_testdata_path,
    ip_tests_enabled,
    testconfig_path,
    migrationdata_path,
    test_users,
    test_groups,
    valid_encodings,
    load_ip_test_data,
    load_molecule_data,
    load_wavefunction_data,
    load_hash_test_data,
    caplog_handler_at_level,
)



from __future__ import annotations

import time
from typing import TYPE_CHECKING

from qcelemental.models import Molecule

from qcportal.record_models import RecordStatusEnum

if TYPE_CHECKING:
    from qcportal import PortalClient


def test_optimization_full_1(fulltest_client: PortalClient):
    mol = Molecule(symbols=["H", "H"], geometry=[[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]])
    meta, ids = fulltest_client.add_optimizations(
        initial_molecules=[mol],
        program="geometric",
        qc_specification={
            "program": "psi4",
            "method": "b3lyp",
            "basis": "6-31g**",
            "keywords": {"maxiter": 100},
        },
        keywords={"maxiter": 30},
    )

    for i in range(120):
        time.sleep(1)
        rec = fulltest_client.get_optimizations(ids[0])
        if rec.status not in [RecordStatusEnum.running, RecordStatusEnum.waiting]:
            break
    else:
        raise RuntimeError("Did not finish calculation in time")

    assert rec.status == RecordStatusEnum.complete


def test_optimization_full_error_1(fulltest_client: PortalClient):
    mol = Molecule(symbols=["H", "H"], geometry=[[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]])
    meta, ids = fulltest_client.add_optimizations(
        initial_molecules=[mol],
        program="geometric",
        qc_specification={
            "program": "psi4",
            "method": "b3lyp",
            "basis": "6-31g**",
            "keywords": {"maxiter": 1},
        },
        keywords={"maxiter": 30},
    )

    for i in range(120):
        time.sleep(1)
        rec = fulltest_client.get_optimizations(ids[0])
        if rec.status not in [RecordStatusEnum.running, RecordStatusEnum.waiting]:
            break
    else:
        raise RuntimeError("Did not finish calculation in time")

    assert rec.status == RecordStatusEnum.error



from __future__ import annotations

import time
from typing import TYPE_CHECKING

from qcportal.molecules import Molecule
from qcportal.record_models import RecordStatusEnum

if TYPE_CHECKING:
    from qcportal import PortalClient


def test_manybody_full_1(fulltest_client: PortalClient):
    molecule = Molecule(
        symbols=["He", "He", "He", "He"],
        geometry=[0.0, 0.0, 0.0, 0.0, 0.0, 3.77945225, 0.0, 3.77945225, 0.0, 0.0, 3.77945225, 3.77945225],
        fragments=[[0], [1], [2], [3]],
    )

    sp_spec = {
        "program": "psi4",
        "driver": "energy",
        "method": "mp2",
        "basis": "aug-cc-pvdz",
        "keywords": {"e_convergence": 1e-10, "d_convergence": 1e-10},
    }

    mb_keywords = {"max_nbody": None, "bsse_correction": "none"}

    meta, ids = fulltest_client.add_manybodys(
        initial_molecules=[molecule], program="manybody", singlepoint_specification=sp_spec, keywords=mb_keywords
    )

    for i in range(240):
        time.sleep(1)
        rec = fulltest_client.get_manybodys(ids[0])
        if rec.status not in [RecordStatusEnum.running, RecordStatusEnum.waiting]:
            break
    else:
        raise RuntimeError("Did not finish calculation in time")

    assert rec.status == RecordStatusEnum.complete


def test_manybody_full_2(fulltest_client: PortalClient):
    molecule = Molecule(
        symbols=["He", "He", "He", "He"],
        geometry=[0.0, 0.0, 0.0, 0.0, 0.0, 3.77945225, 0.0, 3.77945225, 0.0, 0.0, 3.77945225, 3.77945225],
        fragments=[[0], [1], [2], [3]],
    )

    sp_spec = {
        "program": "psi4",
        "driver": "energy",
        "method": "mp2",
        "basis": "aug-cc-pvdz",
        "keywords": {"e_convergence": 1e-10, "d_convergence": 1e-10},
    }

    mb_keywords = {"max_nbody": 2, "bsse_correction": "cp"}

    meta, ids = fulltest_client.add_manybodys(
        initial_molecules=[molecule], program="manybody", singlepoint_specification=sp_spec, keywords=mb_keywords
    )

    for i in range(240):
        time.sleep(1)
        rec = fulltest_client.get_manybodys(ids[0])
        if rec.status not in [RecordStatusEnum.running, RecordStatusEnum.waiting]:
            break
    else:
        raise RuntimeError("Did not finish calculation in time")

    assert rec.status == RecordStatusEnum.complete



from __future__ import annotations

import time
from typing import TYPE_CHECKING

from qcportal.molecules import Molecule
from qcportal.record_models import RecordStatusEnum

if TYPE_CHECKING:
    from qcportal import PortalClient

molecule_H2 = Molecule(
    symbols=["H", "H"],
    geometry=[0.0, 0.0, 0.0, 0.0, 0.0, 1.401],
)

molecule_O2 = Molecule(symbols=["O", "O"], geometry=[0.0, 0.0, 0.0, 0.0, 0.0, 2.2828])

molecule_H2O = Molecule(
    symbols=["O", "H", "H"],
    geometry=[0.0, 0.0, 0.2217, -4.380867762249869e-17, 1.4309, -0.8867, 4.380867762249869e-17, -1.4309, -0.8867],
)


def test_reaction_full_1(fulltest_client: PortalClient):
    sp_spec = {"program": "psi4", "driver": "energy", "method": "b3lyp", "basis": "def2-tzvp", "keywords": {}}

    rxn_keywords = {}

    meta, ids = fulltest_client.add_reactions(
        stoichiometries=[[(-2.0, molecule_H2), (-1.0, molecule_O2), (2.0, molecule_H2O)]],
        program="reaction",
        singlepoint_specification=sp_spec,
        optimization_specification=None,
        keywords=rxn_keywords,
    )

    for i in range(240):
        time.sleep(1)
        rec = fulltest_client.get_reactions(ids[0])
        if rec.status not in [RecordStatusEnum.running, RecordStatusEnum.waiting]:
            break
    else:
        raise RuntimeError("Did not finish calculation in time")

    assert rec.status == RecordStatusEnum.complete


def test_reaction_full_2(fulltest_client: PortalClient):
    sp_spec = {"program": "psi4", "driver": "energy", "method": "b3lyp", "basis": "def2-tzvp", "keywords": {}}

    opt_spec = {
        "program": "geometric",
        "keywords": {},
        "qc_specification": {"program": "psi4", "method": "pbe0", "basis": "sto-3g", "keywords": {}},
    }

    rxn_keywords = {}

    meta, ids = fulltest_client.add_reactions(
        stoichiometries=[[(-2.0, molecule_H2), (-1.0, molecule_O2), (2.0, molecule_H2O)]],
        program="reaction",
        singlepoint_specification=sp_spec,
        optimization_specification=opt_spec,
        keywords=rxn_keywords,
    )

    for i in range(240):
        time.sleep(1)
        rec = fulltest_client.get_reactions(ids[0])
        if rec.status not in [RecordStatusEnum.running, RecordStatusEnum.waiting]:
            break
    else:
        raise RuntimeError("Did not finish calculation in time")

    assert rec.status == RecordStatusEnum.complete


def test_reaction_full_3(fulltest_client: PortalClient):
    opt_spec = {
        "program": "geometric",
        "keywords": {},
        "qc_specification": {
            "program": "psi4",
            "driver": "energy",
            "method": "pbe0",
            "basis": "def2-tzvp",
            "keywords": {},
        },
    }

    rxn_keywords = {}

    meta, ids = fulltest_client.add_reactions(
        stoichiometries=[[(-2.0, molecule_H2), (-1.0, molecule_O2), (2.0, molecule_H2O)]],
        program="reaction",
        singlepoint_specification=None,
        optimization_specification=opt_spec,
        keywords=rxn_keywords,
    )

    for i in range(360):
        time.sleep(1)
        rec = fulltest_client.get_reactions(ids[0])
        if rec.status not in [RecordStatusEnum.running, RecordStatusEnum.waiting]:
            break
    else:
        raise RuntimeError("Did not finish calculation in time")

    assert rec.status == RecordStatusEnum.complete



import json
import lzma
import random

from qcportal.utils import hash_dict

letters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"


def random_key():
    return "".join(random.choices(letters, k=3))


def random_string():
    return "".join(random.choices(letters, k=20))


def random_float():
    return random.random()


def random_int():
    return random.randint(-1000, 1000)


def random_list():
    return [random.random() for _ in range(15)]


def shuffle_dict(d):
    def shuffle_inner(x):
        a, b = x
        if isinstance(b, dict):
            b = shuffle_dict(b)
        return a, b

    d2 = [shuffle_inner(x) for x in d.items()]
    random.shuffle(d2)
    return dict(d2)


generated = {}


def random_dict():
    d = {}
    for i in range(5):
        key = "".join(random.choices(letters, k=3))
        value = "".join(random.choices(letters, k=20))
        d[random_key()] = random_string()
        d[random_key()] = random_int()
        d[random_key()] = random_float()
        d[random_key()] = random_list()


for n in range(10):
    d = {}
    for i in range(5):
        d[random_key()] = random_string()
        d[random_key()] = random_int()
        d[random_key()] = random_float()
        d[random_key()] = random_list()
        d[random_key()] = random_dict()

    h = hash_dict(d)
    generated[h] = shuffle_dict(d)

    # Shuffle it!
    for i in range(10):
        d2 = shuffle_dict(d)
        h2 = hash_dict(d2)
        assert h2 == h


with lzma.open("test_kw_hash.json.xz", "rt") as f:
    existing = json.load(f)

existing.update(generated)

with lzma.open("test_kw_hash.json-new.xz", "wt") as f:
    json.dump(existing, f, indent=2)



"""
Contains testing infrastructure for QCFractal.
"""

# This file is normally named conftest.py, but doing so
# runs into issues with pytest automatically finding this, but also
# having it be found through the pyproject.toml entry point

import pytest


def pytest_addoption(parser):
    """
    Additional PyTest CLI flags to add

    See `pytest_collection_modifyitems` for handling and `pytest_configure` for adding known in-line marks.
    """

    parser.addoption("--runslow", action="store_true", default=False, help="run slow tests")
    parser.addoption("--client-encoding", type=str, default="application/json", help="set client encoding to test")
    parser.addoption(
        "--fractal-uri", type=str, default="snowflake", help="URI of the fractal instance to run full tests against"
    )


def pytest_collection_modifyitems(config, items):
    """
    Handle test triggers based on the CLI flags
    Use decorators:
    @pytest.mark.slow
    """
    runslow = config.getoption("--runslow")

    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords and not runslow:
            item.add_marker(skip_slow)


def pytest_unconfigure(config):
    pass



"""
Pytest fixtures for QCArchive
"""
from __future__ import annotations

import secrets
from typing import Tuple

import pytest

from qcfractal.config import FractalConfig
from qcfractal.db_socket.socket import SQLAlchemySocket
from qcportal import PortalClient
from qcportal.managers import ManagerName
from .helpers import geoip_path, geoip_filename, ip_tests_enabled, test_users
from .testing_classes import QCATestingPostgresServer, QCATestingSnowflake, _activated_manager_programs


@pytest.fixture(scope="session")
def postgres_server(tmp_path_factory):
    """
    A postgres server instance

    This does not contain any databases (except the postgres defaults).

    This is built only once per session, and automatically deleted after the session. It uses
    a pytest-provided session-scoped temporary directory
    """

    db_path = str(tmp_path_factory.mktemp("db"))
    test_pg = QCATestingPostgresServer(db_path)
    yield test_pg


@pytest.fixture(scope="session")
def session_storage_socket(postgres_server):
    """
    A fixture for a storage socket that lasts the entire testing session
    """

    pg_harness = postgres_server.get_new_harness("session_storage")

    # Create a configuration. Since this is mostly just for a storage socket,
    # We can use defaults for almost all, since a flask server, etc, won't be instantiated
    # Also disable connection pooling in the storage socket
    # (which can leave db connections open, causing problems when we go to delete
    # the database)
    cfg_dict = {}
    cfg_dict["base_folder"] = pg_harness.config.base_folder
    cfg_dict["loglevel"] = "DEBUG"
    cfg_dict["database"] = pg_harness.config.dict()
    cfg_dict["database"]["pool_size"] = 0
    cfg_dict["log_access"] = True

    if ip_tests_enabled:
        cfg_dict["geoip2_dir"] = geoip_path
        cfg_dict["geoip2_filename"] = geoip_filename

    cfg_dict["api"] = {"secret_key": secrets.token_urlsafe(32), "jwt_secret_key": secrets.token_urlsafe(32)}
    qcf_config = FractalConfig(**cfg_dict)

    socket = SQLAlchemySocket(qcf_config)

    # Create the template database for use in re-creating the database
    pg_harness.create_template()

    yield socket, pg_harness


@pytest.fixture(scope="function")
def storage_socket(session_storage_socket):
    """
    A fixture for storage socket that lasts a single test function

    The database is deleted then recreated after the test function
    """

    socket, pg_harness = session_storage_socket
    yield socket
    pg_harness.recreate_database()


@pytest.fixture(scope="function")
def session(storage_socket):
    """
    A single sqlalchemy session associated with a storage socket
    """
    with storage_socket.session_scope() as s:
        yield s


@pytest.fixture(scope="session")
def session_snowflake(postgres_server, pytestconfig):
    """
    A QCFractal testing snowflake, existing for the entire session
    """

    pg_harness = postgres_server.get_new_harness("session_snowflake")
    encoding = pytestconfig.getoption("--client-encoding")
    with QCATestingSnowflake(pg_harness, encoding) as snowflake:
        pg_harness.create_template()
        yield snowflake


@pytest.fixture(scope="function")
def snowflake(session_snowflake):
    """
    A QCFractal snowflake used for testing (function-scoped)

    The underlying database is deleted and recreated after each test function
    """

    yield session_snowflake
    session_snowflake.reset()


@pytest.fixture(scope="session")
def session_secure_snowflake(postgres_server, pytestconfig):
    """
    A QCFractal snowflake with authorization/authentication enabled
    """

    pg_harness = postgres_server.get_new_harness("session_secure_snowflake")
    encoding = pytestconfig.getoption("--client-encoding")
    with QCATestingSnowflake(
        pg_harness, encoding, create_users=True, enable_security=True, allow_unauthenticated_read=False
    ) as snowflake:
        pg_harness.create_template()
        yield snowflake


@pytest.fixture(scope="function")
def secure_snowflake(session_secure_snowflake):
    """
    A QCFractal snowflake with authorization/authentication enabled (function-scoped)
    """

    yield session_secure_snowflake
    session_secure_snowflake.reset()


@pytest.fixture(scope="session")
def session_secure_snowflake_allow_read(postgres_server, pytestconfig):
    """
    A QCFractal snowflake with authorization/authentication enabled, but allowing unauthenticated read
    """

    pg_harness = postgres_server.get_new_harness("session_secure_snowflake_allow_read")
    encoding = pytestconfig.getoption("--client-encoding")
    with QCATestingSnowflake(
        pg_harness, encoding, create_users=True, enable_security=True, allow_unauthenticated_read=True
    ) as snowflake:
        pg_harness.create_template()
        yield snowflake


@pytest.fixture(scope="function")
def secure_snowflake_allow_read(session_secure_snowflake_allow_read):
    """
    A QCFractal snowflake with authorization/authentication enabled, but allowing unauthenticated read (function-scoped)
    """

    yield session_secure_snowflake_allow_read
    session_secure_snowflake_allow_read.reset()


@pytest.fixture(scope="function")
def submitter_client(secure_snowflake):
    client = secure_snowflake.client("submit_user", test_users["submit_user"]["pw"])
    yield client


@pytest.fixture(scope="function")
def snowflake_client(snowflake):
    """
    A client connected to a testing snowflake

    This is for a simple snowflake (no security, no compute) because a lot
    of tests will use this. Other tests will need to use a different fixture
    and manually call client() there
    """

    client = snowflake.client()
    yield client


@pytest.fixture(scope="function")
def activated_manager(storage_socket: SQLAlchemySocket) -> Tuple[ManagerName, int]:
    """
    An activated manager, returning only its manager name and id
    """
    mname = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")

    mid = storage_socket.managers.activate(
        name_data=mname,
        manager_version="v2.0",
        username="bill",
        programs=_activated_manager_programs,
        tags=["*"],
    )

    yield mname, mid


@pytest.fixture(scope="function")
def activated_manager_name(activated_manager) -> ManagerName:
    """
    An activated manager, returning only its manager name
    """

    yield activated_manager[0]


@pytest.fixture(scope="function")
def activated_manager_programs(activated_manager) -> ManagerName:
    """
    An activated manager, returning only its programs
    """

    yield _activated_manager_programs


@pytest.fixture(scope="function")
def fulltest_client(pytestconfig):
    """
    A portal client used for full end-to-end tests
    """

    uri = pytestconfig.getoption("--fractal-uri")

    if uri == "snowflake":
        from qcfractal.snowflake import FractalSnowflake

        s = FractalSnowflake()
        yield s.client()
        s.stop()

    else:
        yield PortalClient(address=uri)
